/*Copyright (C) 2009  Annette Bieniusa.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package avl;

import java.util.Arrays;
import java.util.Map;

import avl.AVLTree;
import dstm.*;
import dstm.Runtime;

public class Test {

	//Test parameters
	static int nApplicationThreads = 8;
	
	//private static final int[] items1 = { 20, 10, 30, 80, 50, 90, 1, 2, 3, 4, 5, 9, 8, 7, 6, 999, 0, -1, -2, -3, -4, -5, -6, -7, -8, 11, 23, 111, 256 };
	
	public static void main(String[] args) {	
		GAO<AVLTree<Integer,String>> root = initTree();
		Integer[] items = new Integer[100];
		for(int i=0; i < items.length; i ++) {
			items[i] = new Integer((int) (Math.random() * 1000));
		}
		
		for(int i = 0; i < 10; i++) {
			System.out.println("iteration: " + i);
			execute(root, new InsertFactory(), nApplicationThreads, items);
			checkTree(root, items);
			execute(root, new DeleteFactory(), nApplicationThreads, items);
		}
		showTree(root);
	}
	
	private static <T extends Comparable<T>> void execute(GAO<AVLTree<T,String>> tree, JobFactory<T> factory, int nThreads, T[] items) {
		Thread[] threads = new Thread[nThreads];
		
		int size = items.length / nThreads + (items.length % nThreads == 0 ? 0 : 1);
		
		for(int i = 0; i < nThreads; i++) {
			int from = Math.min(i * size, items.length);
			int to   = Math.min((i+1) * size, items.length);
			Job job  = factory.create(tree, Arrays.copyOfRange(items, from, to));
			threads[i] = new Thread(job);
		}
		
		for(Thread t : threads) {
			t.start();
		}
		
		for(Thread t : threads) {
			try { 
				t.join(); 
			} catch(InterruptedException e) {}
		}
	}
	
	
	private static GAO<AVLTree<Integer, String>> initTree() {
		AVLTree<Integer,String> emptyTree = new AVLTree<Integer,String>();
		GAO<AVLTree<Integer,String>> root = null;
		boolean done = false;
		do {
			try {	
				Transaction trans = Runtime.createTransacation("init");
				root = GAO.create(emptyTree, trans);
				trans.commit();
				done = true;
			} catch ( StmException e ) { 
				Runtime.resetTransaction();
			}
		} while ( !done );
		Runtime.resetTransaction();
		return root;
	}

	private static void showTree(GAO<AVLTree<Integer, String>> root) {
		Transaction trans = Runtime.createTransacation("show");
		Map<Integer,String> map = trans.read(root).getMap(trans);
		System.out.println("tree: " + map);
		trans.commit();
	}
	
	private static void checkTree(GAO<AVLTree<Integer, String>> root, Integer[] items) {
		Transaction trans = Runtime.createTransacation("check");
		for(Integer i : items) {
			if (!trans.read(root).contains(i, trans)) {
				System.out.println("!!! Not in the tree anymore: " + i);
			}
		}
		trans.commit();
	}
}


// Actual tests
//

abstract class Job implements Runnable {
	public Job() {
	}
	
	public void run() {
		boolean done = false;
		do {
			try {
				Transaction trans = Runtime.createTransacation("job");
				performJob(trans);
				trans.commit();
				done = true;
			} catch(StmException e) {
				Runtime.resetTransaction();
			}	
		} while (!done);
		Runtime.resetTransaction();
	}
	
	abstract void performJob(Transaction trans);
}

abstract class JobFactory<T extends Comparable<T>> {
	abstract public Job create(GAO<AVLTree<T,String>> tree, T[] items);
}



class InsertFactory extends JobFactory<Integer> {
	public Job create(final GAO<AVLTree<Integer,String>> tree, final Integer[] items) {
		return new Job() {
			void performJob(Transaction trans) {
				for(Integer i : items) {
					//System.out.println("Insert "+i);
					trans.read(tree).insert(i, i.toString(), trans);
				}
			}
		};
	}
}

class DeleteFactory extends JobFactory<Integer> {
	public Job create(final GAO<AVLTree<Integer,String>> tree, final Integer[] items) {
		return new Job() {
			void performJob(Transaction trans) {
				for(Integer i : items) {
					//System.out.println("Delete "+i);
					trans.read(tree).delete(i, trans);
				}
			}
		};
	}
}
