/*Copyright (C) 2009  Annette Bieniusa, Thomas Fuhrmann.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package dstm;

import java.util.HashSet;
import java.util.Set;

/**
 * The GAO class serves as control instance for globally accessible objects. It
 * provides transactions access to the version history list, and it executes the
 * consensus protocol that determines which of the pending versions shall be
 * selected for the GAO's version history list.
 * 
 * @author Annette Bieniusa <bieniusa@informatik.uni-freiburg.de>
 */
public class GAO<T> {
	
	/**
	 * Current (most recent) version of the GAO.
	 */
	private GAOV<T> current;
	private Set<Transaction> waitingFor = new HashSet<Transaction>();
	/**
	 * Return a version that is the predecessor of the given version. Thereby, a
	 * transaction can obtain a GAO version that matches its check set.
	 * 
	 * @param beforeGaov
	 *            The returned GAO version shall be older as this given version.
	 * 
	 * @return A recent GAO version, but not the head of the version history.
	 * @throws StmReadConflictException
	 */
	protected  GAOV<T> read(Set<GAOV<? extends Object>> readSet) {
		
		GAOV<T> gaov = current.checkForSuccessfulPendingGAOV(readSet);
		
		do {
			if ( gaov == null ) {
				gaov = current;
			} else {
				gaov = gaov.getPredecessor();
			}	  
		} while ( gaov.inconsistentWith(readSet) );
		return gaov;
		
	}
	
	/**
	 * Create a GAO from a local object. The GAO's thread gets started via the
	 * runtime if the creating transaction committed successfully.
	 * 
	 * @param loc
	 *            Local object that we shall turn into a GAO, t transaction
	 *            within which the GAO is created
	 * @return Globally accessible object
	 */
	// TODO Move this method to the transaction as everything happens there
	// locally
	public static <T> GAO<T> create (LOC<T> loc, Transaction t) {
		GAO<T> gao = new GAO<T> ();
		gao.current = (GAOV<T>) loc;
		loc.createFrom(gao,t);
		t.addToCreateSet(gao,loc);
		return gao;
	}

	
	/**
	 * There is one thread associated to each GAO. It checks for new messages in
	 * the message queues and gets the messages processed.
	 */
	
	@SuppressWarnings("unchecked")
	protected void consume(TransactionMessage m) {
				// System.out.println(Thread.currentThread().getId() +" : "+this
				// + "received :" + m);
				switch (m.getType()) {
				case READ :
					consume ((ReadMessage) m);
					break;
				case COMMIT :
					consume ((CommitMessage<T>) m);
					break;
				case VOTINGRESULT: 
					consume ((VotingResultMessage<T>) m);
					break;
				default :
					throw new RuntimeException("GAO " + this.toString() + 
							" received the wrong kind of message");
			}
	}
		
	
	
	
	private void consume(VotingResultMessage<T> m) {
		/*
		 * 1. if failed remove from pendingVersions 2. if success, change
		 * current; set all others to failed 3. else keep collecting and adjust
		 * probabilities if last vote arrives, start randomized voting (again)
		 */
		GAOV<T> parent = m.getGaov().getPredecessor();
		VotingResult vote = m.getVote();
		if (vote == VotingResult.FAILED) {
			
				parent.removeFromSuccessors(m.getGaov());
			
			if (parent == current) {
				waitingFor.remove(m.getSender());
			} else {
				return;
			}
		} else if (vote == VotingResult.SUCCESS) {
			current = m.getGaov();
			waitingFor = new HashSet<Transaction>();

			Set<GAOV<T>> failures = null;
			
				failures = parent.getAllSuccessorsBut(current);
			
			for (GAOV<T> failed : failures) {
				failed.getWrittenBy().enqueue(
						new VoteMessage<T>(this, failed.getWrittenBy(), VoteMessage.Vote.FAILED));
			}
		} else {
			m.getGaov().setProbability(vote.getResult());
			waitingFor.remove(m.getSender());
			/*
			 * String s = ""; for (Transaction t : waitingFor) { s +=
			 * t.toString(); } System.out.println(this + " is waiting for " +
			 * s);
			 */
		}

		/*
		 * All voting messages we were waiting for were received. Hence a new
		 * round of the protocol can start.
		 */
		if (current.hasSuccessors() && waitingFor.isEmpty()) {
			Set<GAOV<T>> pendings = current.getSuccessors();
			// initialize the number of votes we will wait on for the next round
			for (GAOV<T> p : pendings) {
				waitingFor.add(p.getWrittenBy());
			}
			/*
			 * String x = ""; for (Transaction t : waitingFor) { x +=
			 * t.toString(); } System.out.println(this + " starts a new commit
			 * round and is waiting for " + x);
			 */
			GAOV<T> chosen = pickPendingVersion(pendings);

			chosen.getWrittenBy().enqueue(
					new VoteMessage<T>(this, chosen.getWrittenBy(),VoteMessage.Vote.POS));
			pendings.remove(chosen);
			for (GAOV<T> notchosen : pendings) {
				notchosen.getWrittenBy().enqueue(
						new VoteMessage<T>(this, notchosen.getWrittenBy(), VoteMessage.Vote.NEG));
			}

		}

	}

	private GAOV<T> pickPendingVersion(Set<GAOV<T>> pendings) {
		double sum = 0.0;
		double scaled_sum = 0.0;
		
		for (GAOV<T> gaov : pendings) {
			Double prob = gaov.getProbability();
			if (prob != null) {
				sum += prob;
			} 
		}
		
		for (GAOV<T> gaov : pendings) {
			Double prob = gaov.getProbability();
			if (prob == null) {
				gaov.setProbability(prob/pendings.size());
			} else {
				gaov.setProbability(prob * sum/pendings.size());
			}
			scaled_sum += gaov.getProbability();
		}
		
		double choice = scaled_sum * Math.random();
		for ( GAOV<T> gaov : pendings ) {
			scaled_sum -= gaov.getProbability();
			if ( choice >= scaled_sum ) {
				return gaov;
			}
		}
		assert(false);
		return null;

	}

	private void consume(CommitMessage<T> m) {
		/*
		 * 1. check for write conflicts 2. if none, insert into pending version
		 * tree a) if first to commit,notify transaction of pos vote b) else
		 * notify transaction of neg vote
		 */
		GAOV<T> parent = m.getGaov().getPredecessor();
		Transaction sender = m.getSender();
		VoteMessage<T> answer = null;
			if (parent != current) {
				answer = new VoteMessage<T>(this,sender,VoteMessage.Vote.FAILED);
			} else {
				boolean first = (!parent.hasSuccessors());
				parent.addToSuccessors(m.getGaov());
				if (first) {					
					answer = new VoteMessage<T>(this,sender,VoteMessage.Vote.POS);	
				} else {
					answer = new VoteMessage<T>(this,sender,VoteMessage.Vote.NEG);
				}
				waitingFor.add(sender);
			}
		
		sender.enqueue(answer);
	}

	/**
	 * Consumes a ReadMessage. A GAOV which is consistent with the transaction's
	 * readSet so far is found and cloned. The result is then send back.
	 * 
	 * @param m
	 *            ReadMessage
	 */
	private void consume(ReadMessage m) {
		Transaction sender = m.getSender();
		GAOV<T> gaov;
		gaov = read(m.getReadSet());
		LOC<T> loc = (LOC<T>) gaov.read(sender);
		sender.enqueue(new SendGAOVMessage<T>(this,sender,gaov,loc));
	} 
	
}
